import os
import cv2
import glob
import shutil
import random
import numpy as np
from typing import Union

from anonympy.images.utils import pixelated
from anonympy.images.utils import sap_noise
from anonympy.images.utils import find_middle, find_radius


class imAnonymizer(object):
     """
     Initialize an image or a directory as a imAnonymizer object

     Parameters:
     ----------
     path : Union[numpy.ndarray, str]
          cv2.imread object or path string to a folder.
     dst : str, default None
          destination to save the output folder if a string was passed to `path`. If `dst = None` a new
          foulder will be created in the same directory, else in the directory specified. 

     Returns:
     ----------
     imAnonymizer object
     
     Examples
     ----------
     >>> from anonympy.images import imAnonymizer

    Contructing imAnonymizer object by passing an image:

    >>> img = cv2.imread('C://Users/shakhansho/Downloads/image.png')
    >>> anonym = imAnonymizer(img)
     """
     def __init__(self, path, dst = None):
          if os.path.isdir(path):
               self.path = path     
               self._path = True
               self._img = False
          else:
               self.frame = path.copy()
               self._path = False
               self._img = True

          if dst is not None:
               self.dst = dst
               self._dst = True
          else:
               self._dst = False

          self._FACE = cv2.CascadeClassifier(data_file('cascade.xml'))
          self.scaleFactor = 1.1
          self.minNeighbors = 5


     def _face_blur(self, img, kernel = (15,15), shape = 'c', box = None, fname = None):
          '''
          Function to apply Gaussian blur to an image
          '''
          self.detections = self._FACE.detectMultiScale(img,scaleFactor = self.scaleFactor, minNeighbors = self.minNeighbors)
          if self.detections == tuple():
               if self._img:
                    print(f'No Faces were Detected in the Image')
               elif self.path:
                    print(f'No Faces were Detected in the {fname}')
               return None
          else:
               for face in self.detections:
                    x,y,w,h = face
                    
                    noise = cv2.GaussianBlur(img[y:y+h,x:x+w], kernel, cv2.BORDER_DEFAULT)
                    copy = img.copy()
                    
                    if shape == 'c':
                         # circular
                         new = img.copy()
                         new[y:y+h,x:x+w] = noise

                         #mask
                         mask = np.zeros(new.shape[:2], dtype='uint8')
                         # cirlce parameters 
                         cv2.circle(mask, find_middle(x,y,w,h), find_radius(x,y,w,h), 255, -1)
                         #apply
                         copy[mask > 0] = new[mask > 0]
                         
                    elif shape == 'r':
                         # rectangular
                         copy[y:y+h,x:x+w] = noise
                         
                    else:
                         raise Exception('Possible values: `r` (rectangular) and `c` (circular)')
                         
                    
                    if box == 'r':
                         cv2.rectangle(copy, (x,y) ,(x+w,y+h), (255,0,0), 2)
                    elif box == 'c':
                         cv2.circle(copy, find_middle(x,y,w,h), find_radius(x,y,w,h), (255,0,0), 2)
                    elif box is None:
                         pass
                    else:
                         raise Exception('Possible values: `r` (rectangular) and `c` (circular), default `None`')
                         
               return copy


     def face_blur(self, kernel = (15,15), shape = 'c', box = None):
          '''
          Apply Gaussian Blur to the Face
          Based on cv2.GaussianBlur.
          
          Parameters:
          ----------
          kernel : tuple, default (15, 15)
               Gaussian Kernel Size. [height width]. height and width should be odd and can have different values
          shape : str, default 'c'
               Blurring shape. Possible values: `r` (rectangular blurring) and `c` (circular blurring).
          box : str, default None 
               Bounding box. Possible values: `r` (rectangular) and `c` (circular), default `None`.

          Returns:
          ----------
          If an image was passed, a blurred one will be returned. If a path to a folder was passed, a similar folder
          with blurred images will be created. 
          
          Raises:
          ----------
          Exception:
               * if argument for `shape` is something other than `r` or `c` 
               * if argument for `box` is something other that `r`, `c` or None.

          Notes
          ----------
          For the method to work properly, a face has to be present in the image.
          If a path to a folder was passed to `imAnonymizer`, the method will create a new folder `Output`
          with blurred images, preserving the folder structure and image names.
          
          Examples
          ----------
          >>> from anonympy.images import imAnonymizer
          
          Blurring an image

          >>> img = cv2.imread('C://Users/shakhansho/Downloads/image.jpg')
          >>> anonym = imAnonymizer(img)
          >>> blurred = anonym.face_blur()

          Applying Gaussian Blur to all images in a folder

          >>> path = 'C://Users/shakhansho/Downloads/Images'
          >>> anonym = imAnonymizer(path, dst = 'D;//Output')
          >>> anonym.face_blur(shape = 'r', box = 'r')
          '''
          if self._img:
               return self._face_blur(self.frame, kernel = kernel, shape = shape, box = box)
                    
          elif self._path:
               for filepath in glob.iglob(self.path + "/**/*.*", recursive=True):
                    # Ignore non images
                    if not filepath.endswith((".png", ".jpg", ".jpeg")):
                         continue
                    # Process Image
                    img = cv2.imread(filepath)
                    img = self._face_blur(img, shape = shape, box = box, fname = filepath)

                    output_filepath = filepath.replace(os.path.split(self.path)[1], 'Output')
                    output_dir = os.path.dirname(output_filepath)
                    # Ensure the folder exists
                    os.makedirs(output_dir, exist_ok=True)

                    if img is None:
                         pass
                    else:
                         cv2.imwrite(output_filepath, img)
               if self._dst:
                    data_from = self.path.replace(os.path.split(self.path)[1], 'Output')
                    data_to = os.path.join(self.dst, 'Output')
                    shutil.copytree(data_from, data_to, dirs_exist_ok = True)
                    shutil.rmtree(data_from)


     def _face_SaP(self, img, shape = 'c', box = None, fname = None): 
          '''
          Function to apply Salt and Pepper Noise to an Image 
          '''
          self.detections = self._FACE.detectMultiScale(img, scaleFactor = self.scaleFactor, minNeighbors = self.minNeighbors)
          if self.detections == tuple():
               if self._img:
                    print(f'No Faces were Detected in the Image')
               elif self.path:
                    print(f'No Faces were Detected in the {fname}')
               return None
          else:
               for face in self.detections:
                    x,y,w,h = face

                    noise = sap_noise(img[y:y+h,x:x+w])
                    copy = img.copy()

                    if shape == 'c':
                         # circular
                         new = img.copy()
                         new[y:y+h,x:x+w] = noise

                         #mask
                         mask = np.zeros(new.shape[:2], dtype='uint8')
                         # cirlce parameters 
                         cv2.circle(mask, find_middle(x,y,w,h), find_radius(x,y,w,h), 255, -1)

                         #apply
                         copy[mask > 0] = new[mask > 0]
                         
                    elif shape == 'r':
                         # rectangular
                         copy[y:y+h,x:x+w] = noise

                    else:
                         raise Exception('Possible values: `r` (rectangular) and `c` (circular)')
                    
                    if box == 'r':
                         cv2.rectangle(copy, (x,y),(x+w,y+h),(255,0,0),2)
                    elif box == 'c':
                         cv2.circle(copy, find_middle(x,y,w,h), find_radius(x,y,w,h), (255,0,0), 2)
                    elif box == None:
                         pass
                    else:
                         raise Exception('Possible values: `r` (rectangular) and `c` (circular), default `None`')
               return copy


     def face_SaP(self, shape = 'c', box = None):
          '''
          Add Salt and Pepper Noise.
          
          Parameters:
          ----------
          shape : str, default 'c'
               Blurring shape. Possible values: `r` (rectangular blurring) and `c` (circular blurring).
          box : str, default None
               Bounding box. Possible values: `r` (rectangular) and `c` (circular), default `None`.
               
          Returns:
          ----------
          If an image was passed, an image wiht noise will be returned. If a path to a folder was passed, a similar folder
          with noised images will be created. 

          Raises:
          ----------
          Exception:
               * if argument for `shape` is something other than `r` or `c` 
               * if argument for `box` is something other that `r`, `c` or None.

          Notes
          ----------
          For the method to work properly, a face has to be present in the image.
          If a path to a folder was passed to `imAnonymizer`, the method will create a new folder `Output`
          with noise added to the images, preserving the folder structure and image names.

          Examples
          ----------
          >>> from anonympy.images import imAnonymizer
          
          Add noise to an image 

          >>> img = cv2.imread('C://Users/shakhansho/Downloads/image.jpg')
          >>> anonym = imAnonymizer(img)
          >>> noised = anonym.face_SaP(shape = 'c')

          Add noise to all images in a folder 

          >>> path = 'C://Users/shakhansho/Downloads/Images'
          >>> anonym = imAnonymizer(path, dst = 'D;//Output')
          >>> anonym.face_SaP(shape = 'c', box = None)
          '''
          if self._img:
               return self._face_SaP(self.frame, shape = shape, box = box)
                    
          elif self._path:
               for filepath in glob.iglob(self.path + "/**/*.*", recursive=True):
                    # Ignore non images
                    if not filepath.endswith((".png", ".jpg", ".jpeg")):
                         continue
                    # Process Image
                    img = cv2.imread(filepath)
                    img = self._face_SaP(img, shape = shape, box = box, fname = filepath)

                    output_filepath = filepath.replace(os.path.split(self.path)[1], 'Output')
                    output_dir = os.path.dirname(output_filepath)
                    # Ensure the folder exists
                    os.makedirs(output_dir, exist_ok=True)

                    if img is None:
                         pass
                    else:
                         cv2.imwrite(output_filepath, img)
               if self._dst:
                    data_from = self.path.replace(os.path.split(self.path)[1], 'Output')
                    data_to = os.path.join(self.dst, 'Output')
                    shutil.copytree(data_from, data_to, dirs_exist_ok = True)
                    shutil.rmtree(data_from)


     def _face_pixel(self, img, blocks = 20, shape = 'c', box = None, fname = None):
          '''
          '''
          self.detections = self._FACE.detectMultiScale(img, scaleFactor = self.scaleFactor, minNeighbors = self.minNeighbors)
          if self.detections == tuple():
               if self._img:
                    print(f'No Faces were Detected in the Image')
               elif self.path:
                    print(f'No Faces were Detected in the {fname}')
               return None
          else:
               for face in self.detections:
                    x,y,w,h = face

                    noise = pixelated(img[y:y+h,x:x+w], blocks = blocks)
                    copy = img.copy()

                    if shape == 'c':
                         # circular
                         new = img.copy()
                         new[y:y+h,x:x+w] = noise

                         #mask
                         mask = np.zeros(new.shape[:2], dtype='uint8')
                         # cirlce parameters 
                         cv2.circle(mask, find_middle(x,y,w,h), find_radius(x,y,w,h), 255, -1)

                         #apply
                         copy[mask > 0] = new[mask > 0]
                         
                    elif shape == 'r':
                         # rectangular
                         copy[y:y+h,x:x+w] = noise
                         
                    else:
                         raise Exception('Possible values: `r` (rectangular) and `c` (circular)')
                    
                    if box == 'r':
                         cv2.rectangle(copy, (x,y), (x+w,y+h), (255,0,0), 2)
                    elif box == 'c':
                         cv2.circle(copy, find_middle(x,y,w,h), find_radius(x,y,w,h), (255,0,0), 2)
                    elif box is None:
                         pass
                    else:
                         raise Exception('Possible values: `r` (rectangular) and `c` (circular), default `None`')

               return copy


     def face_pixel(self, blocks = 20, shape = 'c', box = None):
          '''
          Add Pixelated Bluring to a Face

          Parameters:
          ----------
          blocks : int, default 20
               face image dimensions are divided into MxN blocks.
          shape : str, default 'c'
               Blurring shape. Possible values: `r` (rectangular blurring) and `c` (circular blurring).
          box : str, default None 
               Bounding box. Possible values: `r` (rectangular) and `c` (circular), default `None`.

          Returns:
          ----------
          If an image was passed, an image wiht pixaled blurred face will be returned. If a path to a folder was passed, a similar folder
          with noised images will be created. 

          Raises:
          ----------
          Exception:
               * if argument for `shape` is something other than `r` or `c` 
               * if argument for `box` is something other that `r`, `c` or None.

          Notes
          ----------
          For the method to work properly, a face has to be present in the image.
          If a path to a folder was passed to `imAnonymizer`, the method will create a new folder `Output`
          with blurred pixaled face, while preserving the folder structure and image names.
          
          Examples
          ----------
          >>> from anonympy.images import imAnonymizer
          
          Blur Face in an image 

          >>> img = cv2.imread('C://Users/shakhansho/Downloads/image.jpg')
          >>> anonym = imAnonymizer(img)
          >>> noised = anonym.face_pixel(shape = 'c')

          Blur all images in a folder 

          >>> path = 'C://Users/shakhansho/Downloads/Images'
          >>> anonym = imAnonymizer(path, dst = 'D;//Output')
          >>> anonym.face_pixel(shape = 'r', box = 'r')
          '''
          if self._img:
               return self._face_pixel(self.frame, blocks = blocks, shape = shape, box = box)
                    
          elif self._path:
               for filepath in glob.iglob(self.path + "/**/*.*", recursive=True):
                    # Ignore non images
                    if not filepath.endswith((".png", ".jpg", ".jpeg")):
                         continue
                    # Process Image
                    img = cv2.imread(filepath)
                    img = self._face_pixel(img, blocks = blocks, shape = shape, box = box, fname = filepath)

                    output_filepath = filepath.replace(os.path.split(self.path)[1], 'Output')
                    output_dir = os.path.dirname(output_filepath)
                    # Ensure the folder exists
                    os.makedirs(output_dir, exist_ok=True)

                    if img is None:
                         pass
                    else:
                         cv2.imwrite(output_filepath, img)

               if self._dst:
                    data_from = self.path.replace(os.path.split(self.path)[1], 'Output')
                    data_to = os.path.join(self.dst, 'Output')
                    shutil.copytree(data_from, data_to, dirs_exist_ok = True)
                    shutil.rmtree(data_from)
          

     def _blur(self, img, method='Gaussian', kernel=(15, 15)):
          '''          
          '''
          if method.lower() == 'gaussian':
               return cv2.GaussianBlur(img, kernel, cv2.BORDER_DEFAULT) 
          elif method.lower() == 'median':
               if type(kernel) == tuple:
                    ksize = kernel[0]
               else:
                    ksize = kernel
               return  cv2.medianBlur(img, ksize)
          elif method.lower() == 'bilateral':
               return  cv2.bilateralFilter(img, *kernel)
          elif method.lower() == 'averaging':
               return cv2.blur(img, kernel)


     def blur(self, method='Gaussian', kernel=(15, 15)):
          '''
          Apply blurring to image
          Based on OpenCV functions.

          Parameters:
          ----------
          method : str, default 'Gaussian'
               Available methods:
                    - Averaging.
                         Example kernel size (30, 30)
                    - Gaussian
                         Kernel (height, width). Height and Width should be odd and can have different values, example (15, 15).                          
                    - Bilateral
                         Example kernel size (30, 150, 150),
                    - Median
                         Example kernel size 11,
          kernel : tuple, default (15, 15)

          Returns:
          ----------
          If an image was passed, a blurred image will be returned. If a path to a folder was passed, a similar folder
          with blurred images will be created. 

          Examples
          ----------
          >>> from anonympy.images import imAnonymizer
          
          Blur an image 

          >>> img = cv2.imread('C://Users/shakhansho/Downloads/image.jpg')
          >>> anonym = imAnonymizer(img)
          >>> blurred = anonym.blur(method = 'gaussian', kernel = (21, 21))

          Blur all images in a folder 

          >>> path = 'C://Users/shakhansho/Downloads/Images'
          >>> anonym = imAnonymizer(path, dst = 'D;//Output')
          >>> anonym.blur(method = 'averaging', kernel = (15, 15))
          '''
          if self._img:
               return self._blur(self.frame, method = method, kernel= kernel)
                    
          elif self._path:
               for filepath in glob.iglob(self.path + "/**/*.*", recursive=True):
                    # Ignore non images
                    if not filepath.endswith((".png", ".jpg", ".jpeg")):
                         continue
                    # Process Image
                    img = cv2.imread(filepath)
                    img = self._blur(img, method= method, kernel= kernel)

                    output_filepath = filepath.replace(os.path.split(self.path)[1], 'Output')
                    output_dir = os.path.dirname(output_filepath)
                    # Ensure the folder exists
                    os.makedirs(output_dir, exist_ok=True)

                    cv2.imwrite(output_filepath, img)
               if self._dst:
                    data_from = self.path.replace(os.path.split(self.path)[1], 'Output')
                    data_to = os.path.join(self.dst, 'Output')
                    shutil.copytree(data_from, data_to, dirs_exist_ok = True)
                    shutil.rmtree(data_from)

def data_file(fname):
          """Return the path to a data file of ours."""
          return os.path.join(os.path.split(__file__)[0], fname)



import cv2
import matplotlib.pyplot as plt
import numpy as np
import imutils

# READING AN IMAGE
img = cv2.imread('car.jpg')
gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

cv2.imshow('image', gray)
##plt.imshow(cv2.cvtColor(gray, cv2.COLOR_BGR2RGB))
##plt.show()



# Apply Filters and Find Edges for localization
bfilter = cv2.bilateralFilter(gray, 11, 17, 17) #Noise reduction
edged = cv2.Canny(bfilter, 30, 200) #Edge detection

cv2.imshow('edged', edged)



# Find Contours and Apply Mask
keypoints = cv2.findContours(edged.copy(), cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
contours = imutils.grab_contours(keypoints)
contours = sorted(contours, key=cv2.contourArea, reverse=True)[:10]

# Loop through and find a rectangle 
location = None
for contour in contours:
    approx = cv2.approxPolyDP(contour, 10, True)
    if len(approx) == 4:
        location = approx
        break

print('Location: ', location)

cv2.circle(img, tuple(location[0][0]), 2, (255, 0,0), 2)
cv2.circle(img, tuple(location[1][0]), 2, (255, 0,0), 2)
cv2.circle(img, tuple(location[2][0]), 2, (255, 0,0), 2)
cv2.circle(img, tuple(location[3][0]), 2, (255, 0,0), 2)
cv2.imshow('final', img)



# Supplementary functions and variables
import random
import numpy as np
import cv2


def find_middle(x, y, w, h) -> tuple:
        '''
        Supple
        '''
        x1, y1 = x, y
        x2, y2 = x + w, y + h
        m1, m2 = int((x1 + x2)/2), int((y1 + y2)/2)
        return m1, m2


def find_radius(x, y, w, h) -> tuple:
        pt1 = (x, y)
        pt2 = (x+w, y+h)

        side_middle = x + w, (y + y + h) / 2
        center = find_middle(x, y, w, h)
        dis = side_middle[0] - center[0]

        return dis


def sap_noise(frame):
    img = frame.copy()
    # Getting the dimensions of the image
    row , col, _ = img.shape
     
    # Randomly pick some pixels in the
    # image for coloring them white
    # Pick a random number between 300 and 10000
    number_of_pixels = random.randint(5000, 10000)
    for i in range(number_of_pixels):
       
        # Pick a random y coordinate
        y_coord=random.randint(0, row - 1)
         
        # Pick a random x coordinate
        x_coord=random.randint(0, col - 1)
         
        # Color that pixel to white
        img[y_coord][x_coord] = 255
         
    # Randomly pick some pixels in
    # the image for coloring them black
    # Pick a random number between 300 and 10000
    number_of_pixels = random.randint(5000 , 10000)
    for i in range(number_of_pixels):
       
        # Pick a random y coordinate
        y_coord=random.randint(0, row - 1)
         
        # Pick a random x coordinate
        x_coord=random.randint(0, col - 1)
         
        # Color that pixel to black
        img[y_coord][x_coord] = 0
         
    return img
     

def pixelated(image,  blocks = 20):
          (h, w) = image.shape[:2]
          xSteps = np.linspace(0, w, blocks + 1, dtype="int")
          ySteps = np.linspace(0, h, blocks + 1, dtype="int")

          for i in range(1, len(ySteps)):
               for j in range(1, len(xSteps)):
                    # compute the starting and ending (x, y)-coordinates
                    # for the current block
                    startX = xSteps[j - 1]
                    startY = ySteps[i - 1]
                    endX = xSteps[j]
                    endY = ySteps[i]
                    # extract the ROI using NumPy array slicing, compute the
                    # mean of the ROI, and then draw a rectangle with the
                    # mean RGB values over the ROI in the original image
                    roi = image[startY:endY, startX:endX]
                    (B, G, R) = [int(x) for x in cv2.mean(roi)[:3]]
                    cv2.rectangle(image, (startX, startY), (endX, endY),(B, G, R), -1)

          return image


def resize(self, new_width=500):
        height, width, _ = self.frame.shape
        ratio = height / width
        new_height = int(ratio * new_width)
        return cv2.resize(self.frame, (new_width, new_height))



from typing import List
from time import time
import pandas as pd


faker_methods = '''A | aba, address, administrative_unit, am_pm, android_platform_token, ascii_company_email, ascii_email, ascii_free_email, ascii_safe_email
B | bank_country, bban, boolean, bothify, bs, building_number
C | cache_pattern, catch_phrase, century, chrome, city, city_prefix, city_suffix, color, color_name, company, company_email, company_suffix, coordinate, country, country_calling_code, country_code, credit_card_expire, credit_card_full, credit_card_number, credit_card_provider, credit_card_security_code, cryptocurrency, cryptocurrency_code, cryptocurrency_name, csv, currency, currency_code, currency_name, currency_symbol, current_country, current_country_code
D | date, date_between, date_between_dates, date_object, date_of_birth, date_this_century, date_this_decade, date_this_month, date_this_year, date_time, date_time_ad, date_time_between, date_time_between_dates, date_time_this_century, date_time_this_decade, date_time_this_month, date_time_this_year, day_of_month, day_of_week, del_arguments, dga, domain_name, domain_word, dsv
E | ean, ean13, ean8, ein, email
F | factories, file_extension, file_name, file_path, firefox, first_name, first_name_female, first_name_male, first_name_nonbinary, fixed_width, format, free_email, free_email_domain, future_date, future_datetime
G | generator_attrs, get_arguments, get_formatter, get_providers
H | hex_color, hexify, hostname, http_method
I | iana_id, iban, image, image_url, internet_explorer, invalid_ssn, ios_platform_token, ipv4, ipv4_network_class, ipv4_private, ipv4_public, ipv6, isbn10, isbn13, iso8601, items, itin
J | job, json
L | language_code, language_name, last_name, last_name_female, last_name_male, last_name_nonbinary, latitude, latlng, lexify, license_plate, linux_platform_token, linux_processor, local_latlng, locale, locales, localized_ean, localized_ean13, localized_ean8, location_on_land, longitude
M | mac_address, mac_platform_token, mac_processor, md5, military_apo, military_dpo, military_ship, military_state, mime_type, month, month_name, msisdn
N | name, name_female, name_male, name_nonbinary, nic_handle, nic_handles, null_boolean, numerify
O | opera
P | paragraph, paragraphs, parse, password, past_date, past_datetime, phone_number, port_number, postalcode, postalcode_in_state, postalcode_plus4, postcode, postcode_in_state, prefix, prefix_female, prefix_male, prefix_nonbinary, pricetag, profile, provider, providers, psv, pybool, pydecimal, pydict, pyfloat, pyint, pyiterable, pylist, pyset, pystr, pystr_format, pystruct, pytimezone, pytuple
R | random, random_choices, random_digit, random_digit_not_null, random_digit_not_null_or_empty, random_digit_or_empty, random_element, random_elements, random_int, random_letter, random_letters, random_lowercase_letter, random_number, random_sample, random_uppercase_letter, randomize_nb_elements, rgb_color, rgb_css_color, ripe_id
S | safari, safe_color_name, safe_domain_name, safe_email, safe_hex_color, secondary_address, seed_instance, seed_locale, sentence, sentences, set_arguments, set_formatter, sha1, sha256, simple_profile, slug, ssn, state, state_abbr, street_address, street_name, street_suffix, suffix, suffix_female, suffix_male, suffix_nonbinary, swift, swift11, swift8
T | tar, text, texts, time, time_delta, time_object, time_series, timezone, tld, tsv
U | unique, unix_device, unix_partition, unix_time, upc_a, upc_e, uri, uri_extension, uri_page, uri_path, url, user_agent, user_name, uuid4
W | weights, windows_platform_token, word, words
Y | year
Z | zipcode, zipcode_in_state, zipcode_plus4'''


_fake_methods = ['aba', 'address', 'administrative_unit', 'am_pm', 'android_platform_token',
                 'ascii_company_email', 'ascii_email', 'ascii_free_email', 'ascii_safe_email', 'bank_country', 'bban',
                 'boolean', 'bothify', 'bs', 'building_number', 'catch_phrase', 'century',
                 'chrome', 'city', 'city_prefix', 'city_suffix', 'color', 'color_name', 'company', 'company_email',
                 'company_suffix', 'coordinate', 'country', 'country_calling_code', 'country_code', 'credit_card_expire',
                 'credit_card_full', 'credit_card_number', 'credit_card_provider', 'credit_card_security_code', 'cryptocurrency',
                 'cryptocurrency_code', 'cryptocurrency_name', 'csv', 'currency', 'currency_code', 'currency_name', 'currency_symbol',
                 'current_country', 'current_country_code', 'date', 'date_between', 'date_between_dates', 'date_object', 'date_of_birth',
                 'date_this_century', 'date_this_decade', 'date_this_month', 'date_this_year', 'date_time', 'date_time_ad', 'date_time_between',
                 'date_time_between_dates', 'date_time_this_century', 'date_time_this_decade', 'date_time_this_month', 'date_time_this_year',
                 'day_of_month', 'day_of_week', 'del_arguments', 'dga', 'domain_name', 'domain_word', 'dsv', 'ean', 'ean13', 'ean8', 'ein', 'email',
                 'factories', 'file_extension', 'file_name', 'file_path', 'firefox', 'first_name', 'first_name_female', 'first_name_male',
                 'first_name_nonbinary', 'fixed_width', 'format', 'free_email', 'free_email_domain', 'future_date', 'future_datetime', 'generator_attrs',
                 'get_arguments', 'get_formatter', 'get_providers', 'hex_color', 'hexify', 'hostname', 'http_method', 'iana_id', 'iban', 'image', 'image_url',
                 'internet_explorer', 'invalid_ssn', 'ios_platform_token', 'ipv4', 'ipv4_network_class', 'ipv4_private', 'ipv4_public', 'ipv6', 'isbn10', 'isbn13',
                 'iso8601', 'items', 'itin', 'job', 'json', 'language_code', 'language_name', 'last_name', 'last_name_female', 'last_name_male', 'last_name_nonbinary',
                 'latitude', 'latlng', 'lexify', 'license_plate', 'linux_platform_token', 'linux_processor', 'local_latlng', 'locale', 'locales', 'localized_ean', 'localized_ean13',
                 'localized_ean8', 'location_on_land', 'longitude', 'mac_address', 'mac_platform_token', 'mac_processor', 'md5', 'military_apo', 'military_dpo', 'military_ship',
                 'military_state', 'mime_type', 'month', 'month_name', 'msisdn', 'name', 'name_female', 'name_male', 'name_nonbinary', 'nic_handle', 'nic_handles', 'null_boolean',
                 'numerify', 'opera', 'paragraph', 'paragraphs', 'parse', 'password', 'past_date', 'past_datetime', 'phone_number', 'port_number', 'postalcode', 'postalcode_in_state',
                 'postalcode_plus4', 'postcode', 'postcode_in_state', 'prefix', 'prefix_female', 'prefix_male', 'prefix_nonbinary', 'pricetag', 'profile', 'provider', 'providers', 'psv',
                 'pybool', 'pydecimal', 'pydict', 'pyfloat', 'pyint', 'pyiterable', 'pylist', 'pyset', 'pystr', 'pystr_format', 'pystruct', 'pytimezone', 'pytuple', 'random', 'random_choices',
                 'random_digit', 'random_digit_not_null', 'random_digit_not_null_or_empty', 'random_digit_or_empty', 'random_element', 'random_elements', 'random_int', 'random_letter', 'random_letters',
                 'random_lowercase_letter', 'random_number', 'random_sample', 'random_uppercase_letter', 'randomize_nb_elements', 'rgb_color', 'rgb_css_color', 'ripe_id', 'safari', 'safe_color_name',
                 'safe_domain_name','safe_email', 'safe_hex_color', 'secondary_address', 'seed', 'seed_instance', 'seed_locale', 'sentence', 'sentences', 'set_arguments', 'set_formatter', 'sha1', 'sha256', 'simple_profile',
                 'slug', 'ssn', 'state', 'state_abbr', 'street_address', 'street_name', 'street_suffix', 'suffix', 'suffix_female', 'suffix_male',
                 'suffix_nonbinary', 'swift', 'swift11', 'swift8', 'tar', 'text', 'texts', 'time', 'time_delta','time_object', 'time_series', 'timezone', 'tld', 'tsv', 'unique',
                 'unix_device', 'unix_partition', 'unix_time', 'upc_a', 'upc_e', 'uri','uri_extension', 'uri_page', 'uri_path', 'url',
                 'user_agent', 'user_name', 'uuid4', 'weights', 'windows_platform_token', 'word', 'words', 'year', 'zipcode', 'zipcode_in_state', 'zipcode_plus4']


av_methods = f'''`numeric`:
        * Perturbation - "numeric_noise"
        * Binning - "numeric_binning"
        * PCA Masking - "numeric_masking"
        * Rounding - "numeric_rounding"

`categorical`:
        * Synthetic Data - "categorical_fake"
        * Synthetic Data Auto - "categorical_fake_auto"
        * Resampling from same Distribution - "categorical_resampling"
        * Tokenazation - "categorical_tokenization"
        * Email Masking - "categorical_email_masking"

`datetime`:
        * Synthetic Date - "datetime_fake"
        * Perturbation - "datetime_noise"

`general`:
        * Drop Column - "column_suppression"
        '''


def get_numeric_columns(df) -> List:
    '''
    Return a subset of the DataFrame's columns which are numeric.

    Returns
    ----------
        List of columns with numeric values 
    '''
    
    return df.select_dtypes('number').columns.tolist()


def get_categorical_columns(df) -> List:
    '''
    Return a subset of the DataFrame's columns which are categorical.

    Returns
    ----------
        List of columns with categorical values 
    '''
    
    return df.select_dtypes(include=['object', 'category']).columns.tolist()


def get_datetime_columns(df) -> List:
    '''
    Return a subset of the DataFrame's columns which are of datetime type. 

    Returns
    ----------
        List of columns with datetime values 
    '''
    
    return df.select_dtypes(include=['datetime']).columns.tolist()


def timer_func(func):
    '''
    This decorator-function shows the execution time of 
    the function 
    '''
    def wrap_func(*args, **kwargs):
        t1 = time()
        result = func(*args, **kwargs)
        t2 = time()
        print(f'Function {func.__name__!r} executed in {(t2-t1):.4f}s')
        return result
    return wrap_func


def fake_methods(letter = None):
    '''
    Print a list of faker's methods
    '''
    if letter == None or letter == 'all':
        print(faker_methods)
    else:
        for line in faker_methods.split('\n'):
            if line[0] == letter.upper():
                print(line[3:])

def available_methods(dtype = None):
    '''
    Print a list of available methods
    '''
    splitted = {'numeric': ['numeric_noise', 'numeric_binning', 'numeric_masking', 'numeric_rounding'],
                'categorical': ['categorical_fake', 'categorical_fake_auto', 'categorical_resampling', 'categorical_tokenization', 'categorical_email_masking'],
                'datetime': ["datetime_fake", "datetime_noise"],
                'general': ['column_suppression']}
    
    if dtype == None or dtype == 'all':
        print(av_methods)
    else:
        print(*splitted[dtype], sep='\t')


def load_dataset(size='small'):
    '''
    Sample dataset for demonstration purposes only

    Returns
    ----------
        df : pd.DataFrame 
    '''
    if size == 'small':
        df = pd.DataFrame({
            "name": ["Bruce", "Tony"],
            "age": [33, 48],
            "birthdate": [pd.Timestamp(1915, 4, 17), pd.Timestamp(1970, 5, 29)],
            "salary": [59234.32, 49324.53],
            "web": ['http://www.alandrosenburgcpapc.co.uk', 'http://www.capgeminiamerica.co.uk'],
            "email": ['josefrazier@owen.com', 'eryan@lewis.com'],
            "ssn": ["343554334", "656564664"]})

        return df
    elif size == 'big':
        df = pd.read_csv('https://raw.githubusercontent.com/ArtLabss/open-data-anonimizer/main/examples/files/new.csv', parse_dates=['birthdate']) 
        return df
